# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen
from tchannel import TChannel, Response
from tchannel.serializer import json, thrift
from tchannel.sync import TChannel as TChannelSync

from yarpc import Request, encoding
from yarpc._future import fail_to


class TChannelInbound(object):

    def __init__(self, tchannel):
        if isinstance(tchannel, TChannelSync):
            raise ValueError(
                "TChannelInbound must use a regular TChannel,"
                "do not use a TChannelSync"
            )
        self._tchannel = tchannel

    def start(self, dispatcher):
        @self._tchannel.register(TChannel.FALLBACK)
        def handler(request):
            procedure = request.endpoint

            yarpc_req = Request(
                caller=request.transport.caller_name,
                service=request.service,
                encoding=request.transport.scheme,
                procedure=procedure,
                body=request.body,
                hostport=self._tchannel.hostport,
            )

            return_future = gen.Future()
            dispatch_future = gen.maybe_future(
                dispatcher.dispatch(yarpc_req),
            )

            # TODO handle headers
            if yarpc_req.encoding == encoding.JSON:
                serializer = json.JsonSerializer()
            else:
                serializer = thrift.ThriftSerializer(None)
            headers = serializer.serialize_header(None)

            @fail_to(return_future)
            def dispatch_done(f):
                result = f.result()
                response = Response(
                    body=result.body,
                    headers=headers,
                    transport=None,  # TODO transport?
                )
                return_future.set_result(response)

            dispatch_future.add_done_callback(dispatch_done)

            return return_future

        if not self._tchannel.is_listening():
            self._tchannel.listen()

    @property
    def hostport(self):
        return self._tchannel.hostport

    @property
    def port(self):
        return int(self.hostport.rsplit(':', 1)[1])



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from functools import wraps

from tornado import gen

from . import THRIFT
from yarpc.request import Request
from yarpc.errors import ValueExpectedError
from yarpc._future import fail_to


class ThriftEncoding(object):

    # TODO based on transport properties, do the following:
    # some outbounds, like tchannel, expect no enveloping on thrift payload
    # outbound = self._rpc.get_outbound(request.service)
    # envelope = getattr(outbound, 'ENVELOPE_THRIFT', True)
    _envelope = False

    NAME = THRIFT

    def __init__(self, rpc):
        self._rpc = rpc

    def __call__(self, request=None, **kwargs):
        if request is None:
            request = Request(**kwargs)

        request.encoding = self.NAME

        thrift_request = request.body
        thrift_module = thrift_request.__thrift_module__
        function_spec = thrift_request.type_spec.function
        service_spec = function_spec.service

        # get service from thrift.load, if not provided
        if request.service is None:
            request.service = thrift_request.__class__.service_name

        request.procedure = _get_procedure_name(
            service=service_spec.name,
            function=function_spec.name,
        )

        # serialize body based on envelope bool
        if self._envelope is True:
            request.body = thrift_module.dumps.message(request.body)
        else:
            request.body = thrift_module.dumps(request.body)

        return_future = gen.Future()
        rpc_call_future = self._rpc.call(request)

        @fail_to(return_future)
        def rpc_call_done(f):
            result = f.result()
            if self._envelope is True:
                thrift_response = thrift_module.loads.message(
                    service_spec.surface,
                    result.body,
                ).body
            else:
                thrift_response = thrift_module.loads(
                    thrift_request.result_type,
                    result.body,
                )
            result.body = _unwrap_body(
                response_spec=thrift_request.result_type.type_spec,
                body=thrift_response,
                procedure=request.procedure
            )
            return_future.set_result(result)

        rpc_call_future.add_done_callback(rpc_call_done)

        return return_future

    def register(self, thrift_service, handler):
        # TODO enable calling with decorator

        procedure = _get_procedure_name(
            service=thrift_service.name,
            function=handler.__name__,
        )
        function = getattr(thrift_service, handler.__name__)

        thrift_request_cls = function.request
        thrift_module = thrift_request_cls.__thrift_module__

        @wraps(handler)
        def wrapped(request):

            # deserialize
            if self._envelope is True:
                request.body = thrift_module.loads.message(
                    thrift_service,
                    request.body,
                ).body
            else:
                request.body = thrift_module.loads(
                    thrift_request_cls,
                    request.body,
                )

            # execute procedure
            response_future = gen.maybe_future(handler(request))

            def on_got_response(f):
                # TODO: f.exception()
                response = f.result()
                # serialize
                response.body = function.response_cls(success=response.body)
                if self._envelope is True:
                    response.body = thrift_module.dumps.message(response.body)
                else:
                    response.body = thrift_module.dumps(response.body)

            response_future.add_done_callback(on_got_response)

            return response_future

        self._rpc.register(procedure, wrapped)


def _get_procedure_name(service, function):
    return '%s::%s' % (service, function)


def _unwrap_body(response_spec, body, procedure):

    # exception - reraise
    for exc_spec in response_spec.exception_specs:
        exc = getattr(body, exc_spec.name)
        if exc is not None:
            raise exc

    # success - non-void
    if response_spec.return_spec is not None:
        if body.success is None:
            raise ValueExpectedError(
                'Expected a value to be returned for %s, '
                'but recieved None - only void procedures can '
                'return None.' % procedure
            )

        return body.success

    # success - void
    else:
        return None



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import json

from tornado import gen
from tornado.web import Application, RequestHandler

from . import behavior

BEHAVIORS = {
    'raw': behavior.raw,
    'json': behavior.json,
    'thrift': behavior.thrift,
}

PASSED = 'passed'
SKIPPED = 'skipped'
FAILED = 'failed'


def start():
    harness = Application([
        (r"/", TestCaseHandler),
    ], debug=True)
    harness.listen(8080, '0.0.0.0')


class TestCaseHandler(RequestHandler):

    def head(self):
        """Crossdock makes a HEAD request to see when client is ready."""
        pass

    @gen.coroutine
    def get(self):
        """Crossdock sends GET requests with query params to initiate test."""
        behavior = self.get_query_argument('behavior')
        server = self.get_query_argument('server')
        transport = self.get_query_argument('transport')

        fn = BEHAVIORS.get(behavior)

        if fn is None:
            self.write(json.dumps([{
                "status": SKIPPED,
                "output": "Not implemented",
            }]))
            return

        try:
            resp = yield gen.maybe_future(fn(server, transport))
            self.write(json.dumps([{
                "status": PASSED,
                "output": resp
            }]))
        except Exception as e:
            self.write(json.dumps([{
                "status": FAILED,
                "output": "%s" % e
            }]))
            return



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from functools import wraps
import json

from tornado import gen

from . import JSON
from yarpc.request import Request
from yarpc._future import fail_to


class JsonEncoding(object):
    # TODO: This object model still feels too complicated. I just want to
    # define a serialize/deserialize method, and the handler I register applies
    # that encoding....

    NAME = JSON

    def __init__(self, rpc):
        self._rpc = rpc

    def __call__(self, request=None, **kwargs):
        if request is None:
            request = Request(**kwargs)

        # serialize
        request.encoding = self.NAME
        request.body = json.dumps(request.body)

        return_future = gen.Future()
        rpc_call_future = self._rpc.call(request)

        @fail_to(return_future)
        def rpc_call_done(f):
            result = f.result()
            result.body = json.loads(result.body)
            return_future.set_result(result)

        rpc_call_future.add_done_callback(rpc_call_done)

        return return_future

    def register(self, procedure, handler):
        @wraps(handler)
        def wrapped(request):

            # deserialize
            request.body = json.loads(request.body)

            return_future = gen.Future()
            handler_future = gen.maybe_future(handler(request))

            @fail_to(return_future)
            def handler_done(f):
                result = f.result()
                result.body = json.dumps(result.body)
                return_future.set_result(result)

            handler_future.add_done_callback(handler_done)

            return return_future

        self._rpc.register(procedure, wrapped)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import os

from tornado import gen

import crossdock
from crossdock import rand, transport
from yarpc import RPC, Request
from yarpc.thrift import load

idl = os.path.join(
    os.path.dirname(crossdock.__file__),
    'Echo.thrift',
)
service = load(idl, service='yarpc-test')


@gen.coroutine
def thrift(server, transport_name):
    rpc = RPC(
        service='client',
        outbounds={
            'yarpc-test': transport.factory(transport_name),
        },
    )

    expected = rand.string(7)

    request = Request(
        service='yarpc-test',  # TODO get this from thrift service
        body=service.Echo.echo(service.Ping(beep=expected)),
        hostport="%s:%s" % (server, transport.port(transport_name)),
    )
    response = yield rpc.thrift(request)

    if response.body.boop != expected:
        raise Exception(
            "expected %s, got %s" % (expected, response.body)
        )

    raise gen.Return("Server said: %s" % response.body.boop)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen
import tornado.web
import tornado.httpserver
import tornado.netutil

from yarpc import Request
from yarpc._future import fail_to
from . import headers


class HTTPInbound(object):

    def __init__(self, port, reuse_port=False, handlers=None):
        self._port = port
        self._reuse_port = reuse_port
        self._handlers = handlers or []

    def start(self, dispatcher):
        self._handlers.append(
            ('.*$', _TornadoRequestHandler, dict(dispatcher=dispatcher)),
        )
        app = tornado.web.Application(
            handlers=self._handlers,
        )
        http_server = tornado.httpserver.HTTPServer(app)

        # the following does the same as http_server.listen(self._port),
        # except with fine grain control over the sockopts
        sockets = tornado.netutil.bind_sockets(
            port=self._port,
            reuse_port=self._reuse_port,
        )
        http_server.add_sockets(sockets)

    @property
    def hostport(self):
        # TODO get pub ip???
        return "localhost:%s" % self._port

    @property
    def port(self):
        return self._port


class _TornadoRequestHandler(tornado.web.RequestHandler):

    def initialize(self, dispatcher):
        self.dispatcher = dispatcher

    def post(self, path=None):

        # from request
        procedure = self.request.headers.get(headers.PROCEDURE)
        request = Request(
            caller=self.request.headers.get(headers.CALLER),
            service=self.request.headers.get(headers.SERVICE),
            encoding=self.request.headers.get(headers.ENCODING),
            procedure=procedure,
            body=self.request.body,
            hostport=self.request.headers.get(headers.HOSTPORT),
        )

        return_future = gen.Future()
        dispatch_future = gen.maybe_future(
            self.dispatcher.dispatch(request),
        )

        @fail_to(return_future)
        def dispatch_done(f):
            result = f.result()
            if result.body is not None:
                self.write(result.body)
            return_future.set_result(result)

        dispatch_future.add_done_callback(dispatch_done)

        return return_future



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from crossdock import rand, transport
from yarpc import Request
from yarpc.sync import RPC


def raw(server, transport_name):
    rpc = RPC(
        service='client',
        outbounds={
            'yarpc-test': transport.factory(transport_name),
        },
    )

    expected = rand.string(7)

    request = Request(
        service='yarpc-test',
        procedure='echo/raw',
        body=expected,
        hostport="%s:%s" % (server, transport.port(transport_name)),
    )
    future = rpc.raw(request)
    response = future.result()

    if response.body != expected:
        raise Exception(
            "expected %s, got %s" % (expected, response.body)
        )

    return "Server said: %s" % response.body



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from threadloop import ThreadLoop

from yarpc.rpc import RPC as AsyncRPC
from yarpc.errors import YARPCError


class RPC(AsyncRPC):
    """YARPC from sync-python; can be used with outbounds only."""

    def __init__(self, service, outbounds=None, threadloop=None):
        super(RPC, self).__init__(
            service=service,
            inbounds=None,
            outbounds=outbounds,
        )
        self._threadloop = threadloop or ThreadLoop()

        # TODO switch to golang-style encoding clients
        self.call = _ThreadLoopCallable(self.call, self._threadloop)
        self.raw = _ThreadLoopCallable(self.raw, self._threadloop)
        self.json = _ThreadLoopCallable(self.json, self._threadloop)
        self.thrift = _ThreadLoopCallable(self.thrift, self._threadloop)

    def register(self, procedure, handler):
        raise SyncRegisterNotSupportedError(
            "Sync is client-only and does not support registration"
        )


class _ThreadLoopCallable(object):

    def __init__(self, fn, threadloop):
        self._fn = fn
        self._threadloop = threadloop

    def __call__(self, *args, **kwargs):
        if not self._threadloop.is_ready():
            self._threadloop.start()
        return self._threadloop.submit(self._fn, *args, **kwargs)


class SyncRegisterNotSupportedError(YARPCError):
    pass



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tchannel import TChannel

from crossdock import transport
from yarpc import RPC
from yarpc.transport import http, tchannel
from . import procedure


def start():
    rpc = RPC(
        service='yarpc-test',
        inbounds=[
            http.HTTPInbound(transport.HTTP_PORT),
            tchannel.TChannelInbound(
                TChannel(
                    name='yarpc-test',  # TODO shouldnt rpc set this somehow?
                    hostport="0.0.0.0:%s" % transport.TCHANNEL_PORT
                ),
            )
        ],
    )
    procedure.raw(rpc)
    procedure.json(rpc)
    procedure.thrift(rpc)
    rpc.start()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from types import ModuleType

import thriftrw

from yarpc.errors import OneWayNotSupportedError


def load(path, service=None):
    # get thriftrw module
    thriftrw_module = thriftrw.load(path)

    # start building a fresh module
    module = ModuleType(thriftrw_module.__name__)

    # attach services
    for service_cls in thriftrw_module.__services__:
        setattr(
            module,
            service_cls.service_spec.name,
            Service(service_cls, module, service),
        )

    # attach types
    for t in thriftrw_module.__types__:
        # __types__ contains primitive type defs in tuples,
        # we only want the types that have __name__
        if hasattr(t, '__name__'):
            setattr(module, t.__name__, t)

    # attach constants
    for k, v in thriftrw_module.__constants__.items():
        setattr(module, k, v)

    # TODO deal with includes...

    return module


class Service(object):

    def __init__(self, cls, module, service_name=None):

        self.service_spec = cls.service_spec

        # attach functions
        for func_spec in self.service_spec.functions:
            setattr(
                self,
                func_spec.name,
                Function(func_spec, self, module, service_name),
            )

    @property
    def name(self):
        """Name of the Thrift service this object represents."""
        return self.service_spec.name


class Function(object):

    def __init__(self, func_spec, service, module, service_name=None):
        self.spec = func_spec
        self.service = service

        self._func = func_spec.surface

        self.request = self._func.request
        self.request.service_name = service_name

        self.response_cls = self._func.response

        self._module = module

    @property
    def procedure(self):
        return '%s::%s' % (self.service.name, self._func.name)

    @property
    def oneway(self):
        return self.spec.oneway

    def __call__(self, *args, **kwargs):
        if self.oneway:
            raise OneWayNotSupportedError(
                'TChannel+Thrift does not currently support oneway '
                'procedures.'
            )
        call_args = self.request(*args, **kwargs)

        return call_args



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tchannel.serializer import json, thrift
from tornado import gen

from yarpc import Response, encoding
from yarpc._future import fail_to


class TChannelOutbound(object):

    def __init__(self, tchannel):
        self._tchannel = tchannel

    def call(self, request):

        # TODO handle headers
        if request.encoding == encoding.JSON:
            serializer = json.JsonSerializer()
        else:
            serializer = thrift.ThriftSerializer(None)
        headers = serializer.serialize_header(None)

        # TODO use tchannel.request once mash lands
        call_args = {
            'scheme': request.encoding,
            'service': request.service,
            'arg1': request.procedure,
            'arg2': headers,
            'arg3': request.body,
            'timeout': None,  # TODO timeout
            'retry_on': None,  # TODO retry_on
            'routing_delegate': None,
            'hostport': request.hostport,
            'shard_key': None,
            'trace': None,  # TODO trace
        }

        return_future = gen.Future()
        tchannel_call_future = self._tchannel.call(**call_args)

        @fail_to(return_future)
        def tchannel_call_done(f):
            result = f.result()
            response = Response(
                body=result.body,
            )
            return_future.set_result(response)

        tchannel_call_future.add_done_callback(tchannel_call_done)

        return return_future



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPRequest

from yarpc import Response
from yarpc._future import fail_to
from . import headers


class HTTPOutbound(object):

    def call(self, request):
        # this needs to be instantiated here to work with threadloop
        client = AsyncHTTPClient()
        http_request = HTTPRequest(
            url='http://%s/' % request.hostport,
            method='POST',
            headers={
                headers.CALLER: request.caller,
                headers.SERVICE: request.service,
                headers.ENCODING: request.encoding,
                headers.PROCEDURE: request.procedure,
                headers.HOSTPORT: request.hostport,
                headers.TTL: '10000',
            },
            body=request.body,
        )

        return_future = gen.Future()
        client_fetch_future = client.fetch(http_request)

        @fail_to(return_future)
        def client_fetch_done(f):
            result = f.result()
            response = Response(
                body=result.body,
            )
            return_future.set_result(response)

        client_fetch_future.add_done_callback(client_fetch_done)

        return return_future



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from tornado import gen

from .dispatcher import Dispatcher
from .encoding import RawEncoding, JsonEncoding, ThriftEncoding


class RPC(object):

    def __init__(self, service, inbounds=None, outbounds=None):
        self._service = service
        self._inbounds = inbounds

        if outbounds is None:
            self._outbounds = {}
        else:
            self._outbounds = outbounds

        self._dispatcher = Dispatcher()

        # TODO encodings should be injectable
        self.raw = RawEncoding(self)
        self.json = JsonEncoding(self)
        self.thrift = ThriftEncoding(self)

    def register(self, procedure, handler):
        self._dispatcher.register(procedure, handler)

    def start(self):
        return [
            gen.maybe_future(inbound.start(self._dispatcher))
            for inbound in self._inbounds
        ]

    def call(self, request):
        assert request.service, "service is required"
        assert request.encoding, "encoding is required"
        assert request.procedure, "procedure is required"

        request.caller = self._service

        outbound = self._outbounds.get(request.service)
        return_future = outbound.call(request)

        return return_future



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re

from setuptools import find_packages, setup

version = None
version_file = 'yarpc/__init__.py'

with open(version_file, 'r') as f:
    for line in f:
        m = re.match(r'^__version__\s*=\s*(["\'])([^"\']+)\1', line)
        if m:
            version = m.group(2)
            break

if not version:
    raise Exception(
        "Could not determine version number from %s" % version_file
    )

with open('README.md') as f:
    long_description = f.read()

setup(
    name='yarpc',
    version=version,
    description=(
        'YARPC for Python',
    ),
    long_description=long_description,
    author='Grayson Koonce',
    author_email='grayson@uber.com',
    url='https://github.com/yarpc/yarpc-python',
    packages=find_packages(exclude=['tests', 'tests.*']),
    license='MIT',
    install_requires=[
        'tornado>=4.2,<5',
        'thriftrw>=1.1,<2',
    ],
    extras_require={
        'sync': ['futures', 'threadloop>=1,<2'],
        'tchannel': ['tchannel>=0.21,<0.22'],
    }
)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import json

from flask import Flask, request

from . import behavior

BEHAVIORS = {
    'raw': behavior.raw,
    'json': behavior.json,
    'thrift': behavior.thrift,
}

PASSED = 'passed'
SKIPPED = 'skipped'
FAILED = 'failed'


app = Flask(__name__)


@app.route('/', methods=['HEAD'])
def ready():
    """Crossdock makes a HEAD request to see when client is ready."""
    return "ok"


@app.route('/', methods=['GET'])
def handle():
    """Crossdock sends GET requests with query params to initiate test."""
    behavior = request.args['behavior']
    server = request.args['server']
    transport = request.args['transport']

    fn = BEHAVIORS.get(behavior)

    if fn is None:
        return json.dumps([{
            "status": SKIPPED,
            "output": "Not implemented",
        }])

    try:
        resp = fn(server, transport)
        return json.dumps([{
            "status": PASSED,
            "output": resp,
        }])
    except Exception as e:
        return json.dumps([{
            "status": FAILED,
            "outout": "%s" % e
        }])


def start():
    """Start yarpc-python-sync test harness."""
    app.run(host='0.0.0.0', port=8080)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from threading import Thread

import tornado.ioloop


class ThreadedServer(object):
    """Start a YARPC server in a background thread."""

    def __init__(self, rpc):
        self._thread = None
        self._ready = False
        self._rpc = rpc

    def start(self):
        assert self._thread is None, 'thread already started'
        self._thread = Thread(target=self._init)
        self._thread.daemon = True
        self._thread.start()
        while not self._ready:
            pass

    def _init(self):
        self.io_loop = tornado.ioloop.IOLoop()
        self.io_loop.make_current()

        self._rpc.start()

        def callback():
            self._ready = True

        self.io_loop.add_callback(callback)
        self.io_loop.start()
